#!/usr/bin/env bash
#
# Copyright 2017, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# install-libcxx.sh - installs a customized version of libcxx
#

set -e

llvm_url=https://github.com/llvm-mirror/llvm.git
libcxxabi_url=https://github.com/llvm-mirror/libcxxabi.git
libcxx_url=https://github.com/pmem/libcxx.git
top=/tmp
install_path=/usr/local/libcxx
release=release_39
release_pmem=release_39_pmem

export CC=clang
export CXX=clang++

cd $top
git clone -b $release --depth 1 $llvm_url
cd $top/llvm
cd $top/llvm/projects
git clone -b $release --depth 1 $libcxxabi_url
git clone -b $release_pmem --depth 1 $libcxx_url
cd $top
mkdir -p build/abi
mkdir -p build/lib
cd $top/build/abi
cmake -DLLVM_PATH=$top/llvm -DCMAKE_INSTALL_PREFIX=$install_path $top/llvm/projects/libcxxabi/
make install
cd $top/build/lib
cmake -DLLVM_PATH=$top/llvm -DLIBCXX_CXX_ABI=libcxxabi -DLIBCXX_CXX_ABI_INCLUDE_PATHS=$top/llvm/projects/libcxxabi/include -DCMAKE_INSTALL_PREFIX=$install_path -DLIBCXX_CXX_ABI_LIBRARY_PATH=$install_path/lib $top/llvm/projects/libcxx
make install
cd $top
rm -rf llvm build
